local _G = getfenv(0);

-- Returns colored text string
local function BoolCol(bool)
	return (bool and "|cff80ff80" or "|cffff8080");
end

--------------------------------------------------------------------------------------------------------
--                                            Create Extras                                           --
--------------------------------------------------------------------------------------------------------

local tipIcon = ItemRefTooltip:CreateTexture(nil,"BACKGROUND");
tipIcon:SetPoint("TOPRIGHT",ItemRefTooltip,"TOPLEFT",0,-2.5);
tipIcon:SetWidth(42);
tipIcon:SetHeight(42);
tipIcon:Hide();

local tipCount = ItemRefTooltip:CreateFontString(nil,"ARTWORK","NumberFontNormal");
tipCount:SetPoint("BOTTOMRIGHT",tipIcon,"BOTTOMRIGHT",-3,3);
tipCount:SetJustifyH("RIGHT");

--------------------------------------------------------------------------------------------------------
--                                         HOOK: SetHyperlink                                         --
--------------------------------------------------------------------------------------------------------

-- HOOK: ItemRefTooltip + GameTooltip: SetHyperlink
local function SetHyperlink_Hook(self,refString)
	local type, id = refString:match("|?H?(%a+):([^:]+)");
	-- Check ID
	if (not id) or (type ~= "achievement" and self ~= ItemRefTooltip) then
		tipIcon:Hide();
		return;
	-- Spells
	elseif (type == "spell") then
		local name, rank, icon, cost, isFunnel, powerType, castTime, minRange, maxRange = GetSpellInfo(id);
		tipCount:SetText("");
		tipIcon:SetTexture(icon);
		tipIcon:Show();
		self:AddLine("Spell ID: "..id..(rank and rank ~= "" and ", "..rank or ""),0.2,0.6,1);
		self:Show();
	-- Items
	elseif (type == "item") then
		local itemStackCount, _, itemTexture = select(8,GetItemInfo(refString));
		tipCount:SetText(itemStackCount and itemStackCount > 1 and itemStackCount or "");
		tipIcon:SetTexture(itemTexture);
		tipIcon:Show();
	-- Achievements
	elseif (type == "achievement") and (self:NumLines() > 0) and (not IsShiftKeyDown()) then
		local guid = refString:match("achievement:[^:]+:([^:]+)");
		-- Get category
		local catId = GetAchievementCategory(id);
		local category, catParent = GetCategoryInfo(catId);
		local catName;
		while (catParent > 0) do
			catName, catParent = GetCategoryInfo(catParent);
			category = catName.." - "..category;
		end
		-- Get Criteria
		local criteria = {};
		local left, right;
		for i = 7, self:NumLines() do
			left = _G[self:GetName().."TextLeft"..i];
			right = _G[self:GetName().."TextRight"..i];
			tinsert(criteria,{ label = left:GetText(), done = left:GetTextColor() < 0.5 });
			if (right:GetText()) then
				tinsert(criteria,{ label = right:GetText(), done = right:GetTextColor() < 0.5 });
			end
		end
		-- Cache Info
		local complete = _G[self:GetName().."TextLeft3"]:GetText();
		local title, _, _, _, _, _, description, _, icon, reward = select(2,GetAchievementInfo(id));
		-- Icon
		if (self == ItemRefTooltip) then
			tipCount:SetText("");
			tipIcon:SetTexture(icon);
			tipIcon:Show();
		end
		-- Rebuild Tip
		self:ClearLines();
		self:AddLine(title);
		self:AddLine("<"..category..">");
		if (reward) then
			self:AddLine(reward,0.2,0.6,1);
		end
		self:AddLine(description,1,1,1,1);
		self:AddLine(BoolCol(complete:match("earned"))..complete);
		if (#criteria > 0) then
			self:AddLine(" ");
			self:AddLine("Achievement Criteria |cffffffff"..#criteria);
			local r1, g1, b1, r2, g2, b2;
			local myDone1, myDone2;
			local isPlayer = (UnitGUID("player"):sub(3) == guid);
			for i = 1, #criteria, 2 do
				r1, g1, b1 = (criteria[i].done and 0.25 or 0.5), (criteria[i].done and 0.75 or 0.5), (criteria[i].done and 0.25 or 0.5);
				if (criteria[i + 1]) then
					r2, g2, b2 = (criteria[i + 1].done and 0.25 or 0.5), (criteria[i + 1].done and 0.75 or 0.5), (criteria[i + 1].done and 0.25 or 0.5);
				end
				if (not isPlayer) then
					myDone1 = select(3,GetAchievementCriteriaInfo(id,i));
					myDone2 = select(3,GetAchievementCriteriaInfo(id,i + 1));
				end
				self:AddDoubleLine((isPlayer and "" or BoolCol(myDone1).."*|r")..criteria[i].label,criteria[i + 1] and criteria[i + 1].label..(isPlayer and "" or BoolCol(myDone2).."*"),r1,g1,b1,r2,g2,b2);
			end
		end
		self:Show();
	-- Other types not supported, hide icon
	else
		tipIcon:Hide();
	end
end

hooksecurefunc(GameTooltip,"SetHyperlink",SetHyperlink_Hook);
hooksecurefunc(ItemRefTooltip,"SetHyperlink",SetHyperlink_Hook);